@@ -1,7 +1,7 @@
 /*
  * libev event processing core, watcher management
  *
- * Copyright (c) 2007,2008 Marc Alexander Lehmann <libev@schmorp.de>
+ * Copyright (c) 2007,2008,2009 Marc Alexander Lehmann <libev@schmorp.de>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without modifica-
@@ -49,6 +49,18 @@ extern "C" {
 #  include "config.h"
 # endif
 
+# if HAVE_CLOCK_SYSCALL
+#  ifndef EV_USE_CLOCK_SYSCALL
+#   define EV_USE_CLOCK_SYSCALL 1
+#   ifndef EV_USE_REALTIME
+#    define EV_USE_REALTIME  0
+#   endif
+#   ifndef EV_USE_MONOTONIC
+#    define EV_USE_MONOTONIC 1
+#   endif
+#  endif
+# endif
+
 # if HAVE_CLOCK_GETTIME
 #  ifndef EV_USE_MONOTONIC
 #   define EV_USE_MONOTONIC 1
@@ -166,6 +178,14 @@ extern "C" {
 
 /* this block tries to deduce configuration from header-defined symbols and defaults */
 
+#ifndef EV_USE_CLOCK_SYSCALL
+# if __linux && __GLIBC__ >= 2
+#  define EV_USE_CLOCK_SYSCALL 1
+# else
+#  define EV_USE_CLOCK_SYSCALL 0
+# endif
+#endif
+
 #ifndef EV_USE_MONOTONIC
 # if defined (_POSIX_MONOTONIC_CLOCK) && _POSIX_MONOTONIC_CLOCK >= 0
 #  define EV_USE_MONOTONIC 1
@@ -288,13 +308,29 @@ extern "C" {
 #endif
 
 #if EV_USE_INOTIFY
+# include <sys/utsname.h>
+# include <sys/statfs.h>
 # include <sys/inotify.h>
+/* some very old inotify.h headers don't have IN_DONT_FOLLOW */
+# ifndef IN_DONT_FOLLOW
+#  undef EV_USE_INOTIFY
+#  define EV_USE_INOTIFY 0
+# endif
 #endif
 
 #if EV_SELECT_IS_WINSOCKET
 # include <winsock.h>
 #endif
 
+/* on linux, we can use a (slow) syscall to avoid a dependency on pthread, */
+/* which makes programs even slower. might work on other unices, too. */
+#if EV_USE_CLOCK_SYSCALL
+# include <syscall.h>
+# define clock_gettime(id, ts) syscall (SYS_clock_gettime, (id), (ts))
+# undef EV_USE_MONOTONIC
+# define EV_USE_MONOTONIC 1
+#endif
+
 #if EV_USE_EVENTFD
 /* our minimum requirement is glibc 2.7 which has the stub, but not the header */
 # include <stdint.h>
@@ -384,7 +420,7 @@ ev_set_syserr_cb (void (*cb)(const char *msg))
 }
 
 static void noinline
-syserr (const char *msg)
+ev_syserr (const char *msg)
 {
   if (!msg)
     msg = "(libev) system error";
@@ -445,6 +481,11 @@ typedef struct
   WL head;
   unsigned char events;
   unsigned char reify;
+  unsigned char emask; /* the epoll backend stores the actual kernel mask in here */
+  unsigned char unused;
+#if EV_USE_EPOLL
+  unsigned int egen;  /* generation counter to counter epoll bugs */
+#endif
 #if EV_SELECT_IS_WINSOCKET
   SOCKET handle;
 #endif
@@ -607,6 +648,9 @@ array_realloc (int elem, void *base, int *cur, int cnt)
   return ev_realloc (base, elem * *cur);
 }
 
+#define array_init_zero(base,count)	\
+  memset ((void *)(base), 0, sizeof (*(base)) * (count))
+
 #define array_needsize(type,base,cur,cnt,init)			\
   if (expect_false ((cnt) > (cur)))				\
     {								\
@@ -659,19 +703,6 @@ queue_events (EV_P_ W *events, int eventcnt, int type)
 
 /*****************************************************************************/
 
-void inline_size
-anfds_init (ANFD *base, int count)
-{
-  while (count--)
-    {
-      base->head   = 0;
-      base->events = EV_NONE;
-      base->reify  = 0;
-
-      ++base;
-    }
-}
-
 void inline_speed
 fd_event (EV_P_ int fd, int revents)
 {
@@ -719,7 +750,7 @@ fd_reify (EV_P)
           #else
             anfd->handle = _get_osfhandle (fd);
           #endif
-          assert (("libev only supports socket fds in this configuration", ioctlsocket (anfd->handle, FIONREAD, &arg) == 0));
+          assert (("libev: only socket fds supported in this configuration", ioctlsocket (anfd->handle, FIONREAD, &arg) == 0));
         }
 #endif
 
@@ -810,6 +841,7 @@ fd_rearm_all (EV_P)
     if (anfds [fd].events)
       {
         anfds [fd].events = 0;
+        anfds [fd].emask  = 0;
         fd_change (EV_A_ fd, EV_IOFDSET | 1);
       }
 }
@@ -971,18 +1003,6 @@ static int signalmax;
 
 static EV_ATOMIC_T gotsig;
 
-void inline_size
-signals_init (ANSIG *base, int count)
-{
-  while (count--)
-    {
-      base->head   = 0;
-      base->gotsig = 0;
-
-      ++base;
-    }
-}
-
 /*****************************************************************************/
 
 void inline_speed
@@ -1013,7 +1033,7 @@ evpipe_init (EV_P)
 #endif
         {
           while (pipe (evpipe))
-            syserr ("(libev) error creating signal/async pipe");
+            ev_syserr ("(libev) error creating signal/async pipe");
 
           fd_intern (evpipe [0]);
           fd_intern (evpipe [1]);
@@ -1113,7 +1133,7 @@ ev_feed_signal_event (EV_P_ int signum)
   WL w;
 
 #if EV_MULTIPLICITY
-  assert (("feeding signal events is only supported in the default loop", loop == ev_default_loop_ptr));
+  assert (("libev: feeding signal events is only supported in the default loop", loop == ev_default_loop_ptr));
 #endif
 
   --signum;
@@ -1252,8 +1272,9 @@ ev_recommended_backends (void)
   flags &= ~EVBACKEND_KQUEUE;
 #endif
 #ifdef __APPLE__
-  // flags &= ~EVBACKEND_KQUEUE; for documentation
-  flags &= ~EVBACKEND_POLL;
+  /* only select works correctly on that "unix-certified" platform */
+  flags &= ~EVBACKEND_KQUEUE; /* horribly broken, even for sockets */
+  flags &= ~EVBACKEND_POLL;   /* poll is based on kqueue from 10.5 onwards */
 #endif
 
   return flags;
@@ -1513,13 +1534,13 @@ ev_loop_fork (EV_P)
 }
 
 #if EV_VERIFY
-void noinline
+static void noinline
 verify_watcher (EV_P_ W w)
 {
-  assert (("watcher has invalid priority", ABSPRI (w) >= 0 && ABSPRI (w) < NUMPRI));
+  assert (("libev: watcher has invalid priority", ABSPRI (w) >= 0 && ABSPRI (w) < NUMPRI));
 
   if (w->pending)
-    assert (("pending watcher not on pending queue", pendings [ABSPRI (w)][w->pending - 1].w == w));
+    assert (("libev: pending watcher not on pending queue", pendings [ABSPRI (w)][w->pending - 1].w == w));
 }
 
 static void noinline
@@ -1529,9 +1550,9 @@ verify_heap (EV_P_ ANHE *heap, int N)
 
   for (i = HEAP0; i < N + HEAP0; ++i)
     {
-      assert (("active index mismatch in heap", ev_active (ANHE_w (heap [i])) == i));
-      assert (("heap condition violated", i == HEAP0 || ANHE_at (heap [HPARENT (i)]) <= ANHE_at (heap [i])));
-      assert (("heap at cache mismatch", ANHE_at (heap [i]) == ev_at (ANHE_w (heap [i]))));
+      assert (("libev: active index mismatch in heap", ev_active (ANHE_w (heap [i])) == i));
+      assert (("libev: heap condition violated", i == HEAP0 || ANHE_at (heap [HPARENT (i)]) <= ANHE_at (heap [i])));
+      assert (("libev: heap at cache mismatch", ANHE_at (heap [i]) == ev_at (ANHE_w (heap [i]))));
 
       verify_watcher (EV_A_ (W)ANHE_w (heap [i]));
     }
@@ -1542,7 +1563,7 @@ array_verify (EV_P_ W *ws, int cnt)
 {
   while (cnt--)
     {
-      assert (("active index mismatch", ev_active (ws [cnt]) == cnt + 1));
+      assert (("libev: active index mismatch", ev_active (ws [cnt]) == cnt + 1));
       verify_watcher (EV_A_ ws [cnt]);
     }
 }
@@ -1559,15 +1580,15 @@ ev_loop_verify (EV_P)
 
   assert (fdchangemax >= fdchangecnt);
   for (i = 0; i < fdchangecnt; ++i)
-    assert (("negative fd in fdchanges", fdchanges [i] >= 0));
+    assert (("libev: negative fd in fdchanges", fdchanges [i] >= 0));
 
   assert (anfdmax >= 0);
   for (i = 0; i < anfdmax; ++i)
     for (w = anfds [i].head; w; w = w->next)
       {
         verify_watcher (EV_A_ (W)w);
-        assert (("inactive fd watcher on anfd list", ev_active (w) == 1));
-        assert (("fd mismatch between watcher and anfd", ((ev_io *)w)->fd == i));
+        assert (("libev: inactive fd watcher on anfd list", ev_active (w) == 1));
+        assert (("libev: fd mismatch between watcher and anfd", ((ev_io *)w)->fd == i));
       }
 
   assert (timermax >= timercnt);
@@ -1654,6 +1675,8 @@ ev_default_destroy (void)
   struct ev_loop *loop = ev_default_loop_ptr;
 #endif
 
+  ev_default_loop_ptr = 0;
+
 #ifndef _WIN32
   ev_ref (EV_A); /* child watcher */
   ev_signal_stop (EV_A_ &childev);
@@ -1669,8 +1692,7 @@ ev_default_fork (void)
   struct ev_loop *loop = ev_default_loop_ptr;
 #endif
 
-  if (backend)
-    postfork = 1; /* must be in line with ev_loop_fork */
+  postfork = 1; /* must be in line with ev_loop_fork */
 }
 
 /*****************************************************************************/
@@ -1693,7 +1715,7 @@ call_pending (EV_P)
 
         if (expect_true (p->w))
           {
-            /*assert (("non-pending watcher on pending list", p->w->pending));*/
+            /*assert (("libev: non-pending watcher on pending list", p->w->pending));*/
 
             p->w->pending = 0;
             EV_CB_INVOKE (p->w, p->events);
@@ -1734,7 +1756,7 @@ timers_reify (EV_P)
     {
       ev_timer *w = (ev_timer *)ANHE_w (timers [HEAP0]);
 
-      /*assert (("inactive timer on timer heap detected", ev_is_active (w)));*/
+      /*assert (("libev: inactive timer on timer heap detected", ev_is_active (w)));*/
 
       /* first reschedule or stop timer */
       if (w->repeat)
@@ -1743,7 +1765,7 @@ timers_reify (EV_P)
           if (ev_at (w) < mn_now)
             ev_at (w) = mn_now;
 
-          assert (("negative ev_timer repeat value found while processing timers", w->repeat > 0.));
+          assert (("libev: negative ev_timer repeat value found while processing timers", w->repeat > 0.));
 
           ANHE_at_cache (timers [HEAP0]);
           downheap (timers, timercnt, HEAP0);
@@ -1766,14 +1788,14 @@ periodics_reify (EV_P)
     {
       ev_periodic *w = (ev_periodic *)ANHE_w (periodics [HEAP0]);
 
-      /*assert (("inactive timer on periodic heap detected", ev_is_active (w)));*/
+      /*assert (("libev: inactive timer on periodic heap detected", ev_is_active (w)));*/
 
       /* first reschedule or stop timer */
       if (w->reschedule_cb)
         {
           ev_at (w) = w->reschedule_cb (w, ev_rt_now);
 
-          assert (("ev_periodic reschedule callback returned time in the past", ev_at (w) >= ev_rt_now));
+          assert (("libev: ev_periodic reschedule callback returned time in the past", ev_at (w) >= ev_rt_now));
 
           ANHE_at_cache (periodics [HEAP0]);
           downheap (periodics, periodiccnt, HEAP0);
@@ -1911,6 +1933,12 @@ ev_unref (EV_P)
   --activecnt;
 }
 
+void
+ev_now_update (EV_P)
+{
+  time_update (EV_A_ 1e100);
+}
+
 static int loop_done;
 
 void
@@ -2128,12 +2156,13 @@ ev_io_start (EV_P_ ev_io *w)
   if (expect_false (ev_is_active (w)))
     return;
 
-  assert (("ev_io_start called with negative fd", fd >= 0));
+  assert (("libev: ev_io_start called with negative fd", fd >= 0));
+  assert (("libev: ev_io start called with illegal event mask", !(w->events & ~(EV_IOFDSET | EV_READ | EV_WRITE))));
 
   EV_FREQUENT_CHECK;
 
   ev_start (EV_A_ (W)w, 1);
-  array_needsize (ANFD, anfds, anfdmax, fd + 1, anfds_init);
+  array_needsize (ANFD, anfds, anfdmax, fd + 1, array_init_zero);
   wlist_add (&anfds[fd].head, (WL)w);
 
   fd_change (EV_A_ fd, w->events & EV_IOFDSET | 1);
@@ -2149,7 +2178,7 @@ ev_io_stop (EV_P_ ev_io *w)
   if (expect_false (!ev_is_active (w)))
     return;
 
-  assert (("ev_io_stop called with illegal fd (must stay constant after start!)", w->fd >= 0 && w->fd < anfdmax));
+  assert (("libev: ev_io_stop called with illegal fd (must stay constant after start!)", w->fd >= 0 && w->fd < anfdmax));
 
   EV_FREQUENT_CHECK;
 
@@ -2169,7 +2198,7 @@ ev_timer_start (EV_P_ ev_timer *w)
 
   ev_at (w) += mn_now;
 
-  assert (("ev_timer_start called with negative timer repeat value", w->repeat >= 0.));
+  assert (("libev: ev_timer_start called with negative timer repeat value", w->repeat >= 0.));
 
   EV_FREQUENT_CHECK;
 
@@ -2182,7 +2211,7 @@ ev_timer_start (EV_P_ ev_timer *w)
 
   EV_FREQUENT_CHECK;
 
-  /*assert (("internal timer heap corruption", timers [ev_active (w)] == (WT)w));*/
+  /*assert (("libev: internal timer heap corruption", timers [ev_active (w)] == (WT)w));*/
 }
 
 void noinline
@@ -2197,7 +2226,7 @@ ev_timer_stop (EV_P_ ev_timer *w)
   {
     int active = ev_active (w);
 
-    assert (("internal timer heap corruption", ANHE_w (timers [active]) == (WT)w));
+    assert (("libev: internal timer heap corruption", ANHE_w (timers [active]) == (WT)w));
 
     --timercnt;
 
@@ -2251,7 +2280,7 @@ ev_periodic_start (EV_P_ ev_periodic *w)
     ev_at (w) = w->reschedule_cb (w, ev_rt_now);
   else if (w->interval)
     {
-      assert (("ev_periodic_start called with negative interval value", w->interval >= 0.));
+      assert (("libev: ev_periodic_start called with negative interval value", w->interval >= 0.));
       /* this formula differs from the one in periodic_reify because we do not always round up */
       ev_at (w) = w->offset + ceil ((ev_rt_now - w->offset) / w->interval) * w->interval;
     }
@@ -2269,7 +2298,7 @@ ev_periodic_start (EV_P_ ev_periodic *w)
 
   EV_FREQUENT_CHECK;
 
-  /*assert (("internal periodic heap corruption", ANHE_w (periodics [ev_active (w)]) == (WT)w));*/
+  /*assert (("libev: internal periodic heap corruption", ANHE_w (periodics [ev_active (w)]) == (WT)w));*/
 }
 
 void noinline
@@ -2284,7 +2313,7 @@ ev_periodic_stop (EV_P_ ev_periodic *w)
   {
     int active = ev_active (w);
 
-    assert (("internal periodic heap corruption", ANHE_w (periodics [active]) == (WT)w));
+    assert (("libev: internal periodic heap corruption", ANHE_w (periodics [active]) == (WT)w));
 
     --periodiccnt;
 
@@ -2317,12 +2346,12 @@ void noinline
 ev_signal_start (EV_P_ ev_signal *w)
 {
 #if EV_MULTIPLICITY
-  assert (("signal watchers are only supported in the default loop", loop == ev_default_loop_ptr));
+  assert (("libev: signal watchers are only supported in the default loop", loop == ev_default_loop_ptr));
 #endif
   if (expect_false (ev_is_active (w)))
     return;
 
-  assert (("ev_signal_start called with illegal signal number", w->signum > 0));
+  assert (("libev: ev_signal_start called with illegal signal number", w->signum > 0));
 
   evpipe_init (EV_A);
 
@@ -2335,7 +2364,7 @@ ev_signal_start (EV_P_ ev_signal *w)
     sigprocmask (SIG_SETMASK, &full, &prev);
 #endif
 
-    array_needsize (ANSIG, signals, signalmax, w->signum, signals_init);
+    array_needsize (ANSIG, signals, signalmax, w->signum, array_init_zero);
 
 #ifndef _WIN32
     sigprocmask (SIG_SETMASK, &prev, 0);
@@ -2383,7 +2412,7 @@ void
 ev_child_start (EV_P_ ev_child *w)
 {
 #if EV_MULTIPLICITY
-  assert (("child watchers are only supported in the default loop", loop == ev_default_loop_ptr));
+  assert (("libev: child watchers are only supported in the default loop", loop == ev_default_loop_ptr));
 #endif
   if (expect_false (ev_is_active (w)))
     return;
@@ -2418,8 +2447,9 @@ ev_child_stop (EV_P_ ev_child *w)
 #  define lstat(a,b) _stati64 (a,b)
 # endif
 
-#define DEF_STAT_INTERVAL 5.0074891
-#define MIN_STAT_INTERVAL 0.1074891
+#define DEF_STAT_INTERVAL  5.0074891
+#define NFS_STAT_INTERVAL 30.1074891 /* for filesystems potentially failing inotify */
+#define MIN_STAT_INTERVAL  0.1074891
 
 static void noinline stat_timer_cb (EV_P_ ev_timer *w_, int revents);
 
@@ -2433,10 +2463,11 @@ infy_add (EV_P_ ev_stat *w)
 
   if (w->wd < 0)
     {
-      ev_timer_start (EV_A_ &w->timer); /* this is not race-free, so we still need to recheck periodically */
+      w->timer.repeat = w->interval ? w->interval : DEF_STAT_INTERVAL;
+      ev_timer_again (EV_A_ &w->timer); /* this is not race-free, so we still need to recheck periodically */
 
       /* monitor some parent directory for speedup hints */
-      /* note that exceeding the hardcoded limit is not a correctness issue, */
+      /* note that exceeding the hardcoded path limit is not a correctness issue, */
       /* but an efficiency issue only */
       if ((errno == ENOENT || errno == EACCES) && strlen (w->path) < 4096)
         {
@@ -2450,8 +2481,8 @@ infy_add (EV_P_ ev_stat *w)
 
               char *pend = strrchr (path, '/');
 
-              if (!pend)
-                break; /* whoops, no '/', complain to your admin */
+              if (!pend || pend == path)
+                break;
 
               *pend = 0;
               w->wd = inotify_add_watch (fs_fd, path, mask);
@@ -2459,11 +2490,28 @@ infy_add (EV_P_ ev_stat *w)
           while (w->wd < 0 && (errno == ENOENT || errno == EACCES));
         }
     }
-  else
-    ev_timer_stop (EV_A_ &w->timer); /* we can watch this in a race-free way */
 
   if (w->wd >= 0)
-    wlist_add (&fs_hash [w->wd & (EV_INOTIFY_HASHSIZE - 1)].head, (WL)w);
+    {
+      wlist_add (&fs_hash [w->wd & (EV_INOTIFY_HASHSIZE - 1)].head, (WL)w);
+
+      /* now local changes will be tracked by inotify, but remote changes won't */
+      /* unless the filesystem it known to be local, we therefore still poll */
+      /* also do poll on <2.6.25, but with normal frequency */
+      struct statfs sfs;
+
+      if (fs_2625 && !statfs (w->path, &sfs))
+        if (sfs.f_type == 0x1373 /* devfs */
+            || sfs.f_type == 0xEF53 /* ext2/3 */
+            || sfs.f_type == 0x3153464a /* jfs */
+            || sfs.f_type == 0x52654973 /* reiser3 */
+            || sfs.f_type == 0x01021994 /* tempfs */
+            || sfs.f_type == 0x58465342 /* xfs */)
+          return;
+
+      w->timer.repeat = w->interval ? w->interval : fs_2625 ? NFS_STAT_INTERVAL : DEF_STAT_INTERVAL;
+      ev_timer_again (EV_A_ &w->timer);
+    }
 }
 
 static void noinline
@@ -2487,7 +2535,7 @@ static void noinline
 infy_wd (EV_P_ int slot, int wd, struct inotify_event *ev)
 {
   if (slot < 0)
-    /* overflow, need to check for all hahs slots */
+    /* overflow, need to check for all hash slots */
     for (slot = 0; slot < EV_INOTIFY_HASHSIZE; ++slot)
       infy_wd (EV_A_ slot, wd, ev);
   else
@@ -2503,6 +2551,7 @@ infy_wd (EV_P_ int slot, int wd, struct inotify_event *ev)
             {
               if (ev->mask & (IN_IGNORED | IN_UNMOUNT | IN_DELETE_SELF))
                 {
+                  wlist_del (&fs_hash [slot & (EV_INOTIFY_HASHSIZE - 1)].head, (WL)w);
                   w->wd = -1;
                   infy_add (EV_A_ w); /* re-add, no matter what */
                 }
@@ -2525,12 +2574,39 @@ infy_cb (EV_P_ ev_io *w, int revents)
     infy_wd (EV_A_ ev->wd, ev->wd, ev);
 }
 
+void inline_size
+check_2625 (EV_P)
+{
+  /* kernels < 2.6.25 are borked
+   * http://www.ussg.indiana.edu/hypermail/linux/kernel/0711.3/1208.html
+   */
+  struct utsname buf;
+  int major, minor, micro;
+
+  if (uname (&buf))
+    return;
+
+  if (sscanf (buf.release, "%d.%d.%d", &major, &minor, &micro) != 3)
+    return;
+
+  if (major < 2
+      || (major == 2 && minor < 6)
+      || (major == 2 && minor == 6 && micro < 25))
+    return;
+
+  fs_2625 = 1;
+}
+
 void inline_size
 infy_init (EV_P)
 {
   if (fs_fd != -2)
     return;
 
+  fs_fd = -1;
+
+  check_2625 (EV_A);
+
   fs_fd = inotify_init ();
 
   if (fs_fd >= 0)
@@ -2567,9 +2643,8 @@ infy_fork (EV_P)
           if (fs_fd >= 0)
             infy_add (EV_A_ w); /* re-add, no matter what */
           else
-            ev_timer_start (EV_A_ &w->timer);
+            ev_timer_again (EV_A_ &w->timer);
         }
-
     }
 }
 
@@ -2615,9 +2690,12 @@ stat_timer_cb (EV_P_ ev_timer *w_, int revents)
     || w->prev.st_ctime != w->attr.st_ctime
   ) {
       #if EV_USE_INOTIFY
-        infy_del (EV_A_ w);
-        infy_add (EV_A_ w);
-        ev_stat_stat (EV_A_ w); /* avoid race... */
+        if (fs_fd >= 0)
+          {
+            infy_del (EV_A_ w);
+            infy_add (EV_A_ w);
+            ev_stat_stat (EV_A_ w); /* avoid race... */
+          }
       #endif
 
       ev_feed_event (EV_A_ w, EV_STAT);
@@ -2630,16 +2708,12 @@ ev_stat_start (EV_P_ ev_stat *w)
   if (expect_false (ev_is_active (w)))
     return;
 
-  /* since we use memcmp, we need to clear any padding data etc. */
-  memset (&w->prev, 0, sizeof (ev_statdata));
-  memset (&w->attr, 0, sizeof (ev_statdata));
-
   ev_stat_stat (EV_A_ w);
 
-  if (w->interval < MIN_STAT_INTERVAL)
-    w->interval = w->interval ? MIN_STAT_INTERVAL : DEF_STAT_INTERVAL;
+  if (w->interval < MIN_STAT_INTERVAL && w->interval)
+    w->interval = MIN_STAT_INTERVAL;
 
-  ev_timer_init (&w->timer, stat_timer_cb, w->interval, w->interval);
+  ev_timer_init (&w->timer, stat_timer_cb, 0., w->interval ? w->interval : DEF_STAT_INTERVAL);
   ev_set_priority (&w->timer, ev_priority (w));
 
 #if EV_USE_INOTIFY
@@ -2649,7 +2723,7 @@ ev_stat_start (EV_P_ ev_stat *w)
     infy_add (EV_A_ w);
   else
 #endif
-    ev_timer_start (EV_A_ &w->timer);
+    ev_timer_again (EV_A_ &w->timer);
 
   ev_start (EV_A_ (W)w, 1);
 
@@ -2829,6 +2903,23 @@ embed_prepare_cb (EV_P_ ev_prepare *prepare, int revents)
   }
 }
 
+static void
+embed_fork_cb (EV_P_ ev_fork *fork_w, int revents)
+{
+  ev_embed *w = (ev_embed *)(((char *)fork_w) - offsetof (ev_embed, fork));
+
+  ev_embed_stop (EV_A_ w);
+
+  {
+    struct ev_loop *loop = w->other;
+
+    ev_loop_fork (EV_A);
+    ev_loop (EV_A_ EVLOOP_NONBLOCK);
+  }
+
+  ev_embed_start (EV_A_ w);
+}
+
 #if 0
 static void
 embed_idle_cb (EV_P_ ev_idle *idle, int revents)
@@ -2845,7 +2936,7 @@ ev_embed_start (EV_P_ ev_embed *w)
 
   {
     struct ev_loop *loop = w->other;
-    assert (("loop to be embedded is not embeddable", backend & ev_embeddable_backends ()));
+    assert (("libev: loop to be embedded is not embeddable", backend & ev_embeddable_backends ()));
     ev_io_init (&w->io, embed_io_cb, backend_fd, EV_READ);
   }
 
@@ -2858,6 +2949,9 @@ ev_embed_start (EV_P_ ev_embed *w)
   ev_set_priority (&w->prepare, EV_MINPRI);
   ev_prepare_start (EV_A_ &w->prepare);
 
+  ev_fork_init (&w->fork, embed_fork_cb);
+  ev_fork_start (EV_A_ &w->fork);
+
   /*ev_idle_init (&w->idle, e,bed_idle_cb);*/
 
   ev_start (EV_A_ (W)w, 1);
@@ -2874,10 +2968,9 @@ ev_embed_stop (EV_P_ ev_embed *w)
 
   EV_FREQUENT_CHECK;
 
-  ev_io_stop (EV_A_ &w->io);
+  ev_io_stop      (EV_A_ &w->io);
   ev_prepare_stop (EV_A_ &w->prepare);
-
-  ev_stop (EV_A_ (W)w);
+  ev_fork_stop    (EV_A_ &w->fork);
 
   EV_FREQUENT_CHECK;
 }
@@ -2984,7 +3077,7 @@ once_cb (EV_P_ struct ev_once *once, int revents)
   void (*cb)(int revents, void *arg) = once->cb;
   void *arg = once->arg;
 
-  ev_io_stop (EV_A_ &once->io);
+  ev_io_stop    (EV_A_ &once->io);
   ev_timer_stop (EV_A_ &once->to);
   ev_free (once);
 
@@ -2994,13 +3087,17 @@ once_cb (EV_P_ struct ev_once *once, int revents)
 static void
 once_cb_io (EV_P_ ev_io *w, int revents)
 {
-  once_cb (EV_A_ (struct ev_once *)(((char *)w) - offsetof (struct ev_once, io)), revents);
+  struct ev_once *once = (struct ev_once *)(((char *)w) - offsetof (struct ev_once, io));
+
+  once_cb (EV_A_ once, revents | ev_clear_pending (EV_A_ &once->to));
 }
 
 static void
 once_cb_to (EV_P_ ev_timer *w, int revents)
 {
-  once_cb (EV_A_ (struct ev_once *)(((char *)w) - offsetof (struct ev_once, to)), revents);
+  struct ev_once *once = (struct ev_once *)(((char *)w) - offsetof (struct ev_once, to));
+
+  once_cb (EV_A_ once, revents | ev_clear_pending (EV_A_ &once->io));
 }
 
 void
